"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const generate_1 = __importDefault(require("./lib/documentation/generate"));
const tsserverlibrary_1 = require("typescript/lib/tsserverlibrary");
const smart_truncate_1 = __importDefault(require("smart-truncate"));
const markdown_table_1 = __importDefault(require("markdown-table"));
const ts_mysql_autocomplete_1 = require("ts-mysql-autocomplete");
const ts_mysql_schema_1 = require("ts-mysql-schema");
const map_severity_1 = require("./lib/map-severity");
const ts_mysql_analyzer_1 = require("ts-mysql-analyzer");
const ts_mysql_parser_1 = __importStar(require("ts-mysql-parser"));
const MAX_LENGTH = 30;
function truncate(str) {
    return smart_truncate_1.default(str, MAX_LENGTH);
}
function padding(amount) {
    return '&nbsp;'.repeat(amount);
}
function createRow(column) {
    const pad = padding(5);
    return [
        truncate(column.name),
        pad,
        truncate(column.sqlType),
        pad,
        truncate(column.tsType),
        pad,
        String(column.optional)
    ];
}
function getKind(type) {
    switch (type) {
        case 'keyword':
            return tsserverlibrary_1.ScriptElementKind.keyword;
        case 'table':
            return tsserverlibrary_1.ScriptElementKind.classElement;
        case 'column':
            return tsserverlibrary_1.ScriptElementKind.memberVariableElement;
    }
}
class MySqlLanguageService {
    constructor({ host, logger, config }) {
        this.config = config;
        this.logger = logger;
        this.host = host;
        const parserOptions = {
            version: this.config.mySQLVersion
        };
        this.autocompleter = new ts_mysql_autocomplete_1.MySQLAutocomplete({ parserOptions });
        this.analyzer = new ts_mysql_analyzer_1.MySQLAnalyzer({ parserOptions });
        const { databaseUri } = config;
        if (!databaseUri) {
            return;
        }
        const mySQLSchema = new ts_mysql_schema_1.MySQLSchema({ uri: databaseUri });
        mySQLSchema
            .getSchema()
            .then(schema => {
            this.schema = schema;
            this.autocompleter = new ts_mysql_autocomplete_1.MySQLAutocomplete({ parserOptions, schema, uppercaseKeywords: true });
            this.analyzer = new ts_mysql_analyzer_1.MySQLAnalyzer({ parserOptions, schema });
            this.onSchemaLoaded();
        })
            .catch(err => {
            this.logger.log('Failed to get schema: ' + err);
        });
    }
    // For testing purposes, tell client the schema has loaded
    onSchemaLoaded() {
        const message = {
            type: 'event',
            event: 'schemaLoadingFinish'
        };
        this.sendResponse(message);
    }
    // For testing purposes, send response to client
    sendResponse(message) {
        const json = JSON.stringify(message);
        const len = Buffer.byteLength(json, 'utf8');
        const formattedMessage = `Content-Length: ${1 + len}\r\n\r\n${json}${this.host.newLine}`;
        this.host.write(formattedMessage);
    }
    createQuickInfo(start, text, docs) {
        return {
            kind: tsserverlibrary_1.ScriptElementKind.string,
            kindModifiers: '',
            textSpan: {
                start,
                length: text.length
            },
            displayParts: [],
            documentation: [
                {
                    kind: '',
                    text: docs
                }
            ],
            tags: []
        };
    }
    hasFileIgnoreComment(context) {
        const contents = this.host.readFile(context.fileName);
        if (!contents) {
            return false;
        }
        const firstLine = contents.split('\n')[0];
        if (firstLine.includes('@ts-mysql-plugin ignore')) {
            return true;
        }
        return false;
    }
    getQuickInfoAtPosition(context, position) {
        var _a;
        if (this.hasFileIgnoreComment(context)) {
            return;
        }
        const offset = context.toOffset(position);
        const parser = new ts_mysql_parser_1.default({ version: this.config.mySQLVersion });
        const statements = parser.splitStatements(context.text);
        const statement = parser.getStatementAtOffset(statements, offset);
        if (!statement) {
            return;
        }
        const result = parser.parse(statement.text);
        const node = parser.getNodeAtOffset(result, offset - statement.start);
        if (!node) {
            return;
        }
        const { start: nodeStart, type } = node;
        const start = statement.start + nodeStart;
        if (type === ts_mysql_parser_1.ReferenceType.KeywordRef) {
            const reference = node;
            const keyword = reference.keyword;
            return this.createQuickInfo(start, keyword, generate_1.default(keyword, 'keyword'));
        }
        if (type === ts_mysql_parser_1.ReferenceType.FunctionRef) {
            const reference = node;
            const fn = reference.function;
            return this.createQuickInfo(start, fn, generate_1.default(fn, 'function'));
        }
        const pad = padding(5);
        const tableHeader = ['Name', pad, 'SQL Type', pad, 'TS Type', pad, 'Optional'];
        const schemaTables = ((_a = this.schema) === null || _a === void 0 ? void 0 : _a.tables) || [];
        if (type === ts_mysql_parser_1.ReferenceType.TableRef) {
            const reference = node;
            const schemaTable = schemaTables.find(t => t.name === reference.table);
            if (schemaTable) {
                const rows = schemaTable.columns.map(column => createRow(column));
                return this.createQuickInfo(start, schemaTable.name, markdown_table_1.default([tableHeader, ...rows]));
            }
        }
        if (type === ts_mysql_parser_1.ReferenceType.ColumnRef) {
            const reference = node;
            const schemaTable = schemaTables.find(t => { var _a; return t.name === ((_a = reference.tableReference) === null || _a === void 0 ? void 0 : _a.table); });
            const schemaColumn = schemaTable === null || schemaTable === void 0 ? void 0 : schemaTable.columns.find(c => c.name === reference.column);
            if (!schemaColumn) {
                return;
            }
            return this.createQuickInfo(start, schemaColumn.name, markdown_table_1.default([tableHeader, createRow(schemaColumn)]));
        }
    }
    getCompletionsAtPosition(context, position) {
        const completionInfo = {
            entries: [],
            isNewIdentifierLocation: false,
            isGlobalCompletion: false,
            isMemberCompletion: false
        };
        if (this.hasFileIgnoreComment(context)) {
            return completionInfo;
        }
        const offset = context.toOffset(position);
        const parser = new ts_mysql_parser_1.default({ version: this.config.mySQLVersion });
        const statements = parser.splitStatements(context.text);
        const statement = parser.getStatementAtOffset(statements, offset);
        if (!statement) {
            return completionInfo;
        }
        const candidates = this.autocompleter.autocomplete(statement.text, offset - statement.start);
        completionInfo.entries = candidates.map(({ text, type }) => {
            return {
                name: text,
                kind: getKind(type),
                kindModifiers: '',
                sortText: text
            };
        });
        return completionInfo;
    }
    getSemanticDiagnostics(context) {
        if (this.hasFileIgnoreComment(context)) {
            return [];
        }
        const diagnostics = this.analyzer.analyze(context.text) || [];
        return diagnostics.map(diagnostic => {
            const substitution = context.getSubstitution(diagnostic.start);
            const stop = substitution ? substitution.oldStop : diagnostic.stop + 1;
            let start = diagnostic.start;
            let length = stop - start;
            // special case to highlight enclosing backticks
            if (!context.text) {
                start = -1;
                length = 2;
            }
            return {
                file: context.node.getSourceFile(),
                source: this.config.pluginName,
                messageText: diagnostic.message,
                category: map_severity_1.mapSeverity(diagnostic.severity),
                code: diagnostic.code,
                length,
                start
            };
        });
    }
}
exports.default = MySqlLanguageService;
//# sourceMappingURL=language-service.js.map